/*
University of Sao Paulo - Institute of Mathematics and Computer Science ( ICMC - USP )
Computer Graphics - SCC 0250
Developer (PAE): Vinicius Ruela Pereira Borges
Profa: Maria Cristina Ferreira de Oliveira

Codigo 9 - Shading
CODE 9 - Shading

> Para compilar este codigo, execute no terminal do Linux - To compile this code, type the following command on Linux Terminal

# gcc shading.c  -lX11 -lglut -lGL -lGLU -lm -o shading

Para executar o codigo, digite - To run this code, type:

# ./shading

*/

//Bibliotecas - Libraries

#define SIZE_GAUSSIAN 40
// Biblioteca de entrada e saida padrao de dados
#include <stdio.h>
// Biblioteca GL: Contem as funcoes padroes do OpenGl
#include <GL/gl.h>
// Biblioteca GLU : Rotinas que utilizam os comandos OpenGL de baixo nivel para executar tarefas
#include <GL/glu.h>
// Biblioteca GLUT : Criacao de janelas e tratamento de eventos de forma independente do sistema operacional
#include <GL/glut.h>
#include <stdlib.h>

// Bibliotecas matemáticas
#include <math.h>

struct vertice
{
    GLfloat x; //ordenada x do vertice
    GLfloat y; //ordenada y do vertice
    GLfloat z; //ordenada z do vertice
};

typedef struct vertice Ponto3D;

struct celula
{
    Ponto3D ponto; //ponto propriamente dito
    Ponto3D normal; //vetor normal
};

typedef struct celula Cell;

// Identificador da janela
int window;

// Estrutura que armazena a Gaussiana
Cell gaussian[SIZE_GAUSSIAN+2][SIZE_GAUSSIAN+2];

// Funcao que prepara a janela onde a cena grafica sera desenhada
void InitGL(int width, int height) {
  
    // Ajusta a cor de fundo para branco
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    
    // Ativa teste de profundidade
    glEnable(GL_DEPTH_TEST);
    
     // Habilita a matriz de modelo        
    glMatrixMode(GL_MODELVIEW);
    
    // Inicializa a matriz de modelo como identidade
    glLoadIdentity();
    
    // Configura a posição de visão da camera
    gluLookAt(1.5, 0.75, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    
    // Habilita a matriz de projeção
    glMatrixMode(GL_PROJECTION);
    
     // Carrega a matriz de identidade
    glLoadIdentity();
    
    // Define a projeção ortogonal
    glOrtho(-2.5, 2.5, -2.5, 2.5, -4.5, 4.5);
    
    // Define a posição e as cores da luz
    GLfloat position[] = {1.0f,1.5f,-0.5f,1.0f};
    GLfloat white[] = {1.0f,1.0f,1.0f,1.0f};
    GLfloat black[] = {0.0f,0.0f,0.0f,1.0f};
    
    // Define a luz puntual da cena: Luz 0
    glLightfv(GL_LIGHT0,GL_POSITION,position);
    glLightfv(GL_LIGHT0,GL_AMBIENT,black);
    glLightfv(GL_LIGHT0,GL_DIFFUSE,white);
    glLightfv(GL_LIGHT0,GL_SPECULAR,white);
    
    // Ativa as rotinas de iluminação
    glEnable(GL_LIGHTING);
    
    // Ativa a luz 0
    glEnable(GL_LIGHT0);
}

// Calcula o produto vetorial
Ponto3D crossproduct(Ponto3D a, Ponto3D b){
  
        Ponto3D prod;
 
	prod.x = a.y*b.z - a.z*b.y;
        prod.y = a.z*b.x - a.x*b.z;
        prod.z = a.x*b.y - a.y*b.x;
        
	return prod;     
}

// Normaliza o vetor normal
Ponto3D normalize(Ponto3D p){
  
        Ponto3D novo;
	
	// Produto escalar
        GLfloat norm = (float) sqrt(p.x*p.x + p.y*p.y + p.z*p.z);
	
        novo.x = p.x/norm;
        novo.y = p.y/norm;
        novo.z = p.z/norm;
	
        return novo;     
}

// Calcula vetor normal
void calculaNormal()
{
     // Contadores
     int i;
     int j;
     
     // Pontos auxiliares
     Ponto3D v1;
     Ponto3D v2;
     Ponto3D v3;
     Ponto3D vec1;
     Ponto3D vec2;
     
     // Calcula a normal de cada face da superfície da Gaussiana
     for(i = 1; i < SIZE_GAUSSIAN-1;i++){
           for( j = 1; j < SIZE_GAUSSIAN-1;j++){
                
                v1 = gaussian[i][j].ponto;
                v2 = gaussian[i][j+1].ponto;
                v3 = gaussian[i+1][j].ponto;
                
		// Subtracao entre tres pontos (v2-v1)x(v1-v3) antes de calcular a normal
                vec1.x = v2.x - v1.x;
                vec1.y = v2.y - v1.y;
                vec1.z = v2.z - v1.z;
                
                vec2.x = v1.x - v3.x;
                vec2.y = v1.y - v3.y;
                vec2.z = v1.z - v3.z;
                
		// Calcula o vetor normal da face e normaliza
                gaussian[i][j].normal = normalize(crossproduct(vec1,vec2));
		
          }
     }
}

// Funcao que desenha a Gaussiana
void criaGaussiana()
{

    // Contadores
    int i;
    int j;
    
    // Variaveis auxiliares
    GLfloat x;
    GLfloat z;
    GLfloat y;
    GLfloat xexp;
    GLfloat zexp;
    
    // Formula da Gaussiana
    for(i = 0; i < SIZE_GAUSSIAN+2;i++){
	for(j = 0; j < SIZE_GAUSSIAN+2;j++){
	    x =  (4.0f * j) / ((float) SIZE_GAUSSIAN + 2) - 2.0f;
	    z =  (4.0f * i) / ((float) SIZE_GAUSSIAN + 2) - 2.0f;
	    
	    xexp = exp(-x * x);
	    zexp = exp(-z * z);
            y = (float) xexp*zexp;
            
	    gaussian[i][j].ponto.x = x;
	    gaussian[i][j].ponto.y = y;
	    gaussian[i][j].ponto.z = z;
	}
    }

}

// Desenha a superficie utilizando renderização flat
void drawSurface_flat()
{
    int i;
    int j;
  
    glPushMatrix();
    
    	// Cria matriz com as células para desenho
	criaGaussiana();
	
	// Calcula a normal de cada face da gaussiana
	calculaNormal();
	
	// Declaracao de variaveis para desenho da superficie
	GLfloat diffuse[] = {0.5f,0.5f,0.8f,1.0f};
	GLfloat specular[] = {1.0f,1.0f,1.0f,1.0f};
	GLfloat shininess = 100.0f;
	
	// Definicao das luzes difusa, especular e coeficiente do material da superficie gaussiana
	glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse);
	glMaterialfv(GL_FRONT,GL_SPECULAR,specular);
	glMaterialf(GL_FRONT,GL_SHININESS,shininess);
	
	// Desenha a superfície
	for (i = 1; i < SIZE_GAUSSIAN - 1; i++) {
	    for (j = 1; j < SIZE_GAUSSIAN - 1; j++) {
	      
	      // Define a normal calculada em cada face da Gaussiana
	      glNormal3f(gaussian[i][j].normal.x,gaussian[i][j].normal.y,gaussian[i][j].normal.z);
	      
		// Desenha cada face
		glBegin(GL_TRIANGLE_STRIP);
		    glVertex3f(gaussian[i+1][j].ponto.x, gaussian[i+1][j].ponto.y, gaussian[i+1][j].ponto.z);
		    glVertex3f(gaussian[i+1][j+1].ponto.x, gaussian[i+1][j+1].ponto.y, gaussian[i+1][j+1].ponto.z);
		    glVertex3f(gaussian[i][j].ponto.x, gaussian[i][j].ponto.y, gaussian[i][j].ponto.z);
		    glVertex3f(gaussian[i][j+1].ponto.x, gaussian[i][j+1].ponto.y, gaussian[i][j+1].ponto.z);
		glEnd();
		
    	    }
	}
	
    glPopMatrix();
}

// Desenha a superficie utilizando renderização flat
void drawSurface_Gouraud()
{
    int i;
    int j;
  
    glPushMatrix();
    
    	// Cria matriz com as células para desenho
	criaGaussiana();
	
	// Calcula a normal de cada face da gaussiana
	calculaNormal();
	
	// Declaracao de variaveis para desenho da superficie
	GLfloat diffuse[] = {0.5f,0.5f,0.8f,1.0f};
	GLfloat specular[] = {1.0f,1.0f,1.0f,1.0f};
	GLfloat shininess = 100.0f;
	
	// Definicao das luzes difusa, especular e coeficiente do material da superficie gaussiana
	glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse);
	glMaterialfv(GL_FRONT,GL_SPECULAR,specular);
	glMaterialf(GL_FRONT,GL_SHININESS,shininess);
	
	// Desenha a superfície
	for (i = 1; i < SIZE_GAUSSIAN - 1; i++) {
	    for (j = 1; j < SIZE_GAUSSIAN - 1; j++) {
	     
		  glBegin(GL_TRIANGLE_STRIP);
		  
		      glNormal3f(gaussian[i+1][j].normal.x,gaussian[i+1][j].normal.y,gaussian[i+1][j].normal.z);
		      glVertex3f(gaussian[i+1][j].ponto.x, gaussian[i+1][j].ponto.y, gaussian[i+1][j].ponto.z);
		      
		      glNormal3f(gaussian[i+1][j+1].normal.x,gaussian[i+1][j+1].normal.y,gaussian[i+1][j+1].normal.z);
		      glVertex3f(gaussian[i+1][j+1].ponto.x, gaussian[i+1][j+1].ponto.y, gaussian[i+1][j+1].ponto.z);
		      
		      glNormal3f(gaussian[i][j].normal.x,gaussian[i][j].normal.y,gaussian[i][j].normal.z);
		      glVertex3f(gaussian[i][j].ponto.x, gaussian[i][j].ponto.y, gaussian[i][j].ponto.z);
		      
		      glNormal3f(gaussian[i][j+1].normal.x,gaussian[i][j+1].normal.y,gaussian[i][j+1].normal.z);
		      glVertex3f(gaussian[i][j+1].ponto.x, gaussian[i][j+1].ponto.y, gaussian[i][j+1].ponto.z);

		glEnd();
		
    	    }
	}
	
    glPopMatrix();
}

// Funcao que desenha a cena grafica na tela
void DrawGLScene() {
  
    // Limpa a tela e o Depth Buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    // Desenha os objetos na cena utilizando o modelo de Flat
    //drawSurface_flat();
    
    // Desenha a superfície Gaussiana utilizando o modelo de Gouraud
    drawSurface_Gouraud();

    // Alterna entre os buffers para cada chamada do desenho
    glutSwapBuffers();
    
    // Processa as rotinas do OpenGL o mais rapido possi­vel
    glFlush();
}


main(int argc, char *argv[])
{

    // Inicializa o Glut
    glutInit(&argc, argv);

    // Ajusta o display para o usuario
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

    // Coordenada inicial da janela
    glutInitWindowPosition(0, 0);

    // Ajusta o tamanho da janela com 640x480
    glutInitWindowSize(800, 800);

    // Nome da Janela
    window = glutCreateWindow("Shading: Flat and Gouraud rendering models");

    // Funcao que prepara a cena grafica na janela com dimensoes 640 x 480
    InitGL(800, 800);

    // Desenha a cena grafica na tela
    glutDisplayFunc(DrawGLScene);
    
    // Diz que a cena eh animada
    glutIdleFunc(DrawGLScene);
    
    // Loop de atualizacao da tela
    glutMainLoop();

    return 0;
}
